/* Name: main.c
 * Project: USBaspLoader
 * Author: Christian Starkjohann
 * Creation Date: 2007-12-08
 * Tabsize: 4
 * Copyright: (c) 2007 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt)
 * This Revision: $Id: main.c 730 2009-03-20 09:05:11Z cs $
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include "usbconfig.h"
#include "usb162.h"
#include "cdc162.h"

#include <avr/boot.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#define LED_ON()    {bit_set(DDRB, PB2); bit_set(PORTB, PB2);}
#define LED_OFF()   {bit_clear(PORTB, PB2); bit_clear(DDRB, PB2);}

#ifndef ASP_LOG
#define asp_debug(ch)
#else
#define ASP_LOG_SIZE	0x80

static uint8_t asp_log[ASP_LOG_SIZE];
static uint8_t asp_log_idx;

void __asp_debug(uint8_t ch) {
	if (asp_log_idx < ASP_LOG_SIZE) {
		asp_log[asp_log_idx++] = ch;

	}
}
#define asp_debug(ch)	__asp_debug(ch)
#endif

static void leaveBootloader() __attribute__((__noreturn__));

#include "bootloaderconfig.h"


#if BOOTLOADER_EMU
#undef FLASHEND
#define FLASHEND	(STEXT-1)
#define EEPROM_SAVE_BASE	0x1f0
#endif


/* ------------------------------------------------------------------------ */

/* Request constants used by USBasp */
#define USBASP_FUNC_CONNECT         1
#define USBASP_FUNC_DISCONNECT      2
#define USBASP_FUNC_TRANSMIT        3
#define USBASP_FUNC_READFLASH       4
#define USBASP_FUNC_ENABLEPROG      5
#define USBASP_FUNC_WRITEFLASH      6
#define USBASP_FUNC_READEEPROM      7
#define USBASP_FUNC_WRITEEEPROM     8
#define USBASP_FUNC_SETLONGADDRESS  9
#define USBASP_FUNC_SETISPSCK       10

/* Block mode data size */
#define USBASP_READBLOCKSIZE   200
#define USBASP_WRITEBLOCKSIZE  200

/* ------------------------------------------------------------------------ */

/* defaults if not in config file: */
#ifndef HAVE_EEPROM_PAGED_ACCESS
#   define HAVE_EEPROM_PAGED_ACCESS 0
#endif
#ifndef HAVE_EEPROM_BYTE_ACCESS
#   define HAVE_EEPROM_BYTE_ACCESS  0
#endif
#ifndef BOOTLOADER_CAN_EXIT
#   define  BOOTLOADER_CAN_EXIT     0
#endif

/* allow compatibility with avrusbboot's bootloaderconfig.h: */
#ifdef BOOTLOADER_INIT
#   define bootLoaderInit()         BOOTLOADER_INIT
#   define bootLoaderExit()
#endif
#ifdef BOOTLOADER_CONDITION
#   define bootLoaderCondition()    BOOTLOADER_CONDITION
#else
#endif

/* device compatibility: */
#ifndef GICR    /* ATMega*8 don't have GICR, use MCUCR instead */
#   define GICR     MCUCR
#endif

/* ------------------------------------------------------------------------ */
#if (FLASHEND) > 0xffff /* we need long addressing */
#   define CURRENT_ADDRESS  currentAddress.l
#   define addr_t           uint32_t
#else
#   define CURRENT_ADDRESS  currentAddress.w[0]
#   define addr_t           uint16_t
#endif

typedef union longConverter{
    addr_t  l;
    uint16_t  w[sizeof(addr_t)/2];
    uint8_t   b[sizeof(addr_t)];
}longConverter_t;

static uint8_t            requestBootLoaderExit;
static longConverter_t  currentAddress; /* in bytes */
static uint8_t            bytesRemaining;
static uint8_t            isLastPage;
static uint8_t            currentRequest;

#if (HAVE_FUSE_ACCESS == 0)
static const uint8_t  signatureBytes[4] = {
#ifdef SIGNATURE_BYTES
    SIGNATURE_BYTES
#elif defined (__AVR_ATmega8__) || defined (__AVR_ATmega8HVA__)
    0x1e, 0x93, 0x07, 0
#elif defined (__AVR_ATmega48__) || defined (__AVR_ATmega48P__)
    0x1e, 0x92, 0x05, 0
#elif defined (__AVR_ATmega88__) || defined (__AVR_ATmega88P__)
    0x1e, 0x93, 0x0a, 0
#elif defined (__AVR_ATmega168__) || defined (__AVR_ATmega168P__)
    0x1e, 0x94, 0x06, 0
#elif defined (__AVR_ATmega328P__)
    0x1e, 0x95, 0x0f, 0
#elif defined (__AVR_AT90USB162__)
    0x1e, 0x94, 0x82, 0
#else
#   error "Device signature is not known, please edit main.c!"
#endif
};
#endif

/* ------------------------------------------------------------------------ */

static void (*nullVector)(void) __attribute__((__noreturn__));

static void leaveBootloader()
{
    //DBG1(0x01, 0, 0);
    bootLoaderExit();
    cli();
#if 0
    GICR = (1 << IVCE);     /* enable change of interrupt vectors */
    GICR = (0 << IVSEL);    /* move interrupts to application flash section */
/* We must go through a global function pointer variable instead of writing
 *  ((void (*)(void))0)();
 * because the compiler optimizes a constant 0 to "rcall 0" which is not
 * handled correctly by the assembler.
 */
#endif
    nullVector();
}

/* ------------------------------------------------------------------------ */
#if HAVE_FUSE_ACCESS
static inline uint8_t read_fuse(uint16_t addr) {
	SPMCSR |= (1<<SPMEN) | (1<<BLBSET);
	return __LPM(addr);
}

static inline uint8_t read_signature(uint16_t addr) {
	SPMCSR |= (1<<SPMEN) | (1<<SIGRD);
	return  __LPM(addr);
}
#endif

uint8_t asp_setup(uint8_t *in, uint8_t *out)
{
uint8_t           len = 0;
uint8_t           i;
uint8_t		  request = in[USB_RQ_REQ]; 
asp_debug(in[USB_RQ_TYPE]);
asp_debug(request);
    if(request == USBASP_FUNC_TRANSMIT){   /* emulate parts of ISP protocol */
        uint8_t ispcmd = in[USB_RQ_VALUE_L];
        uint8_t isparg = in[USB_RQ_VALUE_H];
        uint8_t rval = 0;
	uint16_t address;
//asp_debug(ispcmd);
        address = ((uint16_t)(in[USB_RQ_VALUE_H])<<8 ) | in[USB_RQ_INDEX_L];
        if(ispcmd == 0x30){        /* read signature */
            rval = in[USB_RQ_INDEX_L] & 3;
#if HAVE_FUSE_ACCESS
	    if (rval < 3) {
		    rval <<= 1;
	    } else {
		    rval = 1;
	    }
            cli();
            rval = read_signature(rval);
	    sei();
#else
            rval = signatureBytes[rval];
#endif
#if HAVE_EEPROM_BYTE_ACCESS
        }else if(ispcmd == 0xa0){  /* read EEPROM byte */
            rval = eeprom_read_byte((void *)address);
        }else if(ispcmd == 0xc0){  /* write EEPROM byte */
            eeprom_write_byte((void *)address, in[USB_RQ_INDEX_H]);
#endif
#if HAVE_FLASH_BYTE_ACCESS
        }else if(ispcmd == 0x20 || ispcmd == 0x28){
            address <<= 1;
	    if (ispcmd == 0x28) {  /* read FLASH upper */
		    address += 1;
	    }
#if BOOTLOADER_EMU
	    if (address <= 3) {
            	rval = eeprom_read_byte((void *)address+ EEPROM_SAVE_BASE);
	    } else 
#endif
	    {
            	rval = pgm_read_byte((void *)address);
	    }
#endif
#if HAVE_FUSE_ACCESS
        }else if((ispcmd & 0xf7) == 0x50){
            cli();
	    if (ispcmd == 0x50 && isparg == 0x00) {  /* read FUSE lower */
		    rval = 0;
	    } else if (ispcmd == 0x50 && isparg == 0x08) {  /* read FUSE ext */
		    rval = 2;
	    } else if (ispcmd == 0x58 && isparg == 0x00) {  /* read LOCKBIT */
		    rval = 1;
	    } else if (ispcmd == 0x58 && isparg == 0x08) {  /* read FUSE upper */
		    rval = 3;
	    }
	    rval = read_fuse(rval);
            sei();
#endif
#if HAVE_CHIP_ERASE
        }else if(ispcmd == 0xac && isparg == 0x80){  /* chip erase */
            addr_t addr;
            for(addr = 0; addr < FLASHEND + 1 - 2048; addr += SPM_PAGESIZE) {
                /* wait and erase page */
                //DBG1(0x33, 0, 0);
#   ifndef NO_FLASH_WRITE
                boot_spm_busy_wait();
                cli();
                boot_page_erase(addr);
                sei();
#   endif
            }
#endif
        }else{
            /* ignore all others, return default value == 0 */
        }
        out[3] = rval;
        len = 4;
    }else if(request == USBASP_FUNC_ENABLEPROG || request == USBASP_FUNC_SETISPSCK){
        out[0] = 0;
        len = 1;
    }else if(request >= USBASP_FUNC_READFLASH && request <= USBASP_FUNC_SETLONGADDRESS){
        currentAddress.b[0] = in[USB_RQ_VALUE_L];
        currentAddress.b[1] = in[USB_RQ_VALUE_H];
	len = in[USB_RQ_LENGTH_L];
        if(request == USBASP_FUNC_SETLONGADDRESS){
#if (FLASHEND) > 0xffff
            currentAddress.b[2] = in[USB_RQ_INDEX_L];
            currentAddress.b[3] = in[USB_RQ_INDEX_H];
#endif
	    len = 0;
        }
	else if(request == USBASP_FUNC_READEEPROM || request == USBASP_FUNC_READFLASH) {
    		for(i = 0; i < len; i++){
		    if (request == USBASP_FUNC_READEEPROM) {
            		*out++ = eeprom_read_byte((void *)currentAddress.w[0]);
#if BOOTLOADER_EMU
		    } else if (CURRENT_ADDRESS <= 3L) {
            		    *out++ = eeprom_read_byte(
					(void *)(currentAddress.w[0]+EEPROM_SAVE_BASE));
#endif
		    } else {
            		    *out++ = pgm_read_byte((void *)CURRENT_ADDRESS);
		    }
        	    CURRENT_ADDRESS++;
		}
    		return len;
	} else {
            /* if(in[USB_RQ_REQ] == USBASP_FUNC_WRITEFLASH) only evaluated during writeFlash anyway */
            isLastPage = in[USB_RQ_INDEX_H] & 0x02;
            currentRequest = request;
	    bytesRemaining = len;
            len = 0; // NO ERROR @ OUT
	}
#if BOOTLOADER_CAN_EXIT
    }else if(request == USBASP_FUNC_DISCONNECT){
        requestBootLoaderExit = 1;      /* allow proper shutdown/close of connection */
#endif
    }else{
        /* ignore: USBASP_FUNC_CONNECT */
    }
    return len;
}

uint8_t asp_out(uint8_t *data, uint8_t len)
{
uint8_t   isLast;

asp_debug(len);
asp_debug(bytesRemaining);
    //DBG1(0x31, (void *)&currentAddress.l, 4);
    if(len > bytesRemaining)
        len = bytesRemaining;
    bytesRemaining -= len;
    isLast = (bytesRemaining == 0);
    if(currentRequest >= USBASP_FUNC_READEEPROM){ // USBASP_FUNC_WRITEEEPROM
        uint8_t i;
        for(i = 0; i < len; i++){
            eeprom_write_byte((void *)(currentAddress.w[0]++), *data++);
        }
    }else{ // USBASP_FUNC_WRITEFLASH
        uint8_t i;
        for(i = 0; i < len;){
#if !HAVE_CHIP_ERASE
            if((currentAddress.w[0] & (SPM_PAGESIZE - 1)) == 0){    /* if page start: erase */
                //DBG1(0x33, 0, 0);
#   ifndef NO_FLASH_WRITE
                cli();
                boot_page_erase(CURRENT_ADDRESS);   /* erase page */
                sei();
                boot_spm_busy_wait();               /* wait until page is erased */
#   endif
            }
#endif
            i += 2;
            //DBG1(0x32, 0, 0);
#if BOOTLOADER_EMU
	    if (CURRENT_ADDRESS <= 3L ) {
            	    eeprom_write_word((void *)(EEPROM_SAVE_BASE + CURRENT_ADDRESS), *(short *)data);
		    *(short *)data = __LPM_word(STEXT+CURRENT_ADDRESS);
	    }
#endif
            cli();
            boot_page_fill(CURRENT_ADDRESS, *(short *)data);
            sei();
            CURRENT_ADDRESS += 2;
            data += 2;
            /* write page when we cross page boundary or we have the last partial page */
            if((currentAddress.w[0] & (SPM_PAGESIZE - 1)) == 0 || (isLast && i >= len && isLastPage)){
                //DBG1(0x34, 0, 0);
#ifndef NO_FLASH_WRITE
                cli();
                boot_page_write(CURRENT_ADDRESS - 2);
                sei();
                boot_spm_busy_wait();
                cli();
                boot_rww_enable();
                sei();
#endif
            }
        }
        //DBG1(0x35, (void *)&currentAddress.l, 4);
    }
    return isLast?(uint8_t)255:0; 
}

/* ------------------------------------------------------------------------ */

#ifdef CDC_DEBUG
#define usbInit	usbcdc_init
#define usbPoll usbcdc_poll
#endif

#ifndef ZZ_USBASP
static inline void asp_init() {
#ifdef CDC_DEBUG
	usbInit();
#else
	usb_vendor_setup = asp_setup;
	usb_user_out = asp_out;
	usbInit();
#endif
	sei();
	while (!usb_connected) {
		wdr();					// $BDI2C(B
		usbPoll();
	}
}

int main(void)
{ 
    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = 0;

    bit_set(REGCR, REGDIS); // disable internal regulator

    /* initialize  */
    wdt_disable();      /* main app may have enabled watchdog */
    bootLoaderInit();
    //DBG1(0x00, 0, 0);
#if 0
#ifndef NO_FLASH_WRITE
    GICR = (1 << IVCE);  /* enable change of interrupt vectors */
    GICR = (1 << IVSEL); /* move interrupts to boot flash section */
#endif
#endif
#if 0
    if(bootLoaderCondition()){
#if BOOTLOADER_CAN_EXIT
        uint8_t i = 0, j = 0;
#endif
        asp_init();
        do{
	    wdr();					// $BDI2C(B
            usbPoll();
#if BOOTLOADER_CAN_EXIT
            if(requestBootLoaderExit){
                if(--i == 0){
                    if(--j == 0)
                        break;
                }
            }
#endif
        }while(bootLoaderCondition());  /* main event loop */
    }
#else
        asp_init();
	for (;;) {
	    wdr();					// $BDI2C(B
            usbPoll();
	}
#endif
    leaveBootloader();
    return 0;
}

void callback_set_baud(long baud) {
	//Baud = baud;
}
#endif // ZZ_USBASP

void asp_print_status() {
#ifdef ASP_LOG
	uint8_t i,j,d;
	uint16_t offset = 0;
	usbcdc_put_pstr(PSTR("asp "));
	//usbcdc_putc('0' + asp_state);
	//usbcdc_put_pstr(PSTR(" err:"));
	//usbcdc_put_hex( last_err_cmd );
	usbcdc_put_pstr(PSTR("\r\n"));
	for(i=0; i<ASP_LOG_SIZE/16; i++) {
		d = (offset >> 8) & 0xff;
		usbcdc_put_hex(d);
		d = offset & 0xff;
		usbcdc_put_hex(d);
		usbcdc_putc(':');
		for (j=0; j<16; j++) {
			usbcdc_putc(' ');
			d = asp_log[offset];
			usbcdc_put_hex(d);
			offset++;
		}
		usbcdc_put_pstr(PSTR("\r\n"));
	}
	asp_log_idx = 0;
	memset(asp_log,0,ASP_LOG_SIZE);
#endif
}
