/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )

#include "usbconfig.h"
#include "usb162.h"
#include "cdc162.h"
#include "serflash.h"
#include "msc162.h"

#define DEV_VCC_ON()	{\
	bit_clear(PORTC,PC6);\
	bit_set(DDRC,PC6);\
}

#define LCD_LED_ON()	{\
	bit_clear(PORTB,PB7);\
	bit_set(DDRB,PB7);\
}

static uint32_t flash_offset;

#ifdef EP_TX
void dump_serflash() {
	uint8_t i,j,d;
	serflash_close();
	serflash_open(SERFLASH_READ,flash_offset);
	for(i=0; i<16; i++) {
		d = (flash_offset >> 16) & 0xff;
		usbcdc_put_hex(d);
		d = (flash_offset >> 8) & 0xff;
		usbcdc_put_hex(d);
		d = flash_offset & 0xff;
		usbcdc_put_hex(d);
		usbcdc_putc(':');
		for (j=0; j<16; j++) {
			usbcdc_putc(' ');
			d = serflash_proc(0);
			usbcdc_put_hex(d);
		}
		flash_offset += 16;
		usbcdc_put_pstr(PSTR("\r\n"));
	}
	serflash_close();
}

#ifdef EP_MS_OUT
extern uint8_t mmc_read_16(uint8_t cmd);
extern uint8_t mmc_buf[16];

void mmc_dump(uint8_t id) {
	int8_t i;
	for (i=0; i<16; i++) {
		usbcdc_putc(' ');
		usbcdc_put_hex(mmc_buf[i]);
	}
}
#endif
#endif

uint8_t read_fuse(uint16_t addr) {
	SPMCSR |= (1<<SPMEN) | (1<<BLBSET);
	return __LPM(addr);
}

uint8_t read_signature(uint16_t addr) {
	SPMCSR |= (1<<SPMEN) | (1<<SIGRD);
	return __LPM(addr);
}

#ifdef ZZ_USBASP
extern uint8_t asp_setup(uint8_t *in, uint8_t *out);
extern uint8_t asp_out(uint8_t *data, uint8_t len);
#endif

int main(void)
{
    uint8_t ch,d;

    /* set system_clock pre scaler 1/8 => 1/1 */
    CLKPR = (1<<CLKPCE);
    CLKPR = 0;

    bit_set(REGCR, REGDIS); // disable internal regulator 

    bit_set(DDRB, PB2);
    bit_set(PORTB, PB2); // LED ON

    init_spi2();
    DEV_VCC_ON();

    usb_vendor_setup = asp_setup;
    usb_user_out = asp_out;
    usbcdc_init();

    sei();

#ifdef EP_MS_OUT
    msc_init();
#endif
    for(;;) {
#ifdef EP_MS_OUT
        msc_poll();
#endif
#ifdef EP_TX
        usbcdc_poll();
	if (usbcdc_can_getc(1)) {
	    ch = usbcdc_getc();
	    switch (ch) {
#ifdef EP_MS_OUT
	    case '1':
		mmc_init();
		mmc_dump(1);
		break;
	    case '2':
		mmc_read_16(9);
		mmc_dump(2);
		break;
	    case '3':
		mmc_read_16(10);
		mmc_dump(3);
	    case 'x':
		msc_print_status();
		break;
#endif
#ifdef ZZ_USBASP
	    case 'x':
		asp_print_status();
		break;
#endif
	    case 'c':
	    	usbcdc_put_pstr(PSTR("clear pointer:\r\n"));
		flash_offset = 0;
		break;
	    case 'r':
	    	usbcdc_put_pstr(PSTR("status:"));
		msc_lock(MSC_OWNER_PROG);
		spi2_enable();
		spi2_write(SERFLASH_RDSR);
		d = spi2_read();
		spi2_disable();
	    	usbcdc_put_hex(d);
	    	usbcdc_put_pstr(PSTR("\r\n"));
		msc_unlock();
		break;
	    case 'S':
	    	usbcdc_put_pstr(PSTR("signature:"));
		usbcdc_put_hex(read_signature(0));
	    	usbcdc_putc(' ');
		usbcdc_put_hex(read_signature(2));
	    	usbcdc_putc(' ');
		usbcdc_put_hex(read_signature(4));
	    	usbcdc_putc(' ');
		usbcdc_put_hex(read_signature(1)); // CALIB
	    	usbcdc_put_pstr(PSTR("\r\n"));
		break;
	    case 'F':
	    	usbcdc_put_pstr(PSTR("fuses:"));
		usbcdc_put_hex(read_fuse(1)); // Lock bit
	    	usbcdc_putc(' ');
		usbcdc_put_hex(read_fuse(0)); // FUSE Lower
	    	usbcdc_putc(' ');
		usbcdc_put_hex(read_fuse(3)); // FUSE upper
	    	usbcdc_putc(' ');
		usbcdc_put_hex(read_fuse(2)); // FUSE ext
	    	usbcdc_put_pstr(PSTR("\r\n"));
		break;
	    case 's':
	    	usbcdc_put_pstr(PSTR("signature read:"));
		msc_lock(MSC_OWNER_PROG);
		serflash_init();

	    	usbcdc_put_hex(serflash_signature[0]);
	    	usbcdc_putc(' ');
	    	usbcdc_put_hex(serflash_signature[1]);
	    	usbcdc_putc(' ');
	    	usbcdc_put_hex(serflash_signature[2]);
	    	usbcdc_put_pstr(PSTR("\r\n"));
		msc_unlock();
		break;
	    case 'd':
		msc_lock(MSC_OWNER_PROG);
		dump_serflash();
		msc_unlock();
		break;
	    case 'e':
		msc_lock(MSC_OWNER_PROG);
#ifdef SUPPORT_ERASE4K
		serflash_erase_4k(flash_offset);
#else
		serflash_erase_64k(flash_offset);
#endif
		msc_unlock();
		break;
	    case 'D':
		{
			uint16_t x;
			for (x=0; x<8192; x++) {
				msc_lock(MSC_OWNER_PROG);
				dump_serflash();
				msc_unlock();
			}
		}
		break;
	    case '\r':
	    	usbcdc_putc(ch);
	    	usbcdc_putc('\n');
		break;
	    default: 
	    	usbcdc_putc(ch);
	    }
	}
#endif
    }
    return 0;
}

void callback_set_baud(long baud) {
	//Baud = baud;
}

