/* Name: cdc162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-9
 * Tabsize: 8
 * Copyright: (c) 2007 by Koji Suzuki (suz@nmj2.sakura.ne.jp)
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#include "usbconfig.h"
#include "usb162.h"
#include "cdc162.h"
#include "msc162.h"

#ifdef EP_MS_OUT
/* ---- */
void usbmsc_poll() {
	UENUM = EP_MS_OUT;
	if (bit_is_set(UEINTX,RXOUTI)) {
		bit_clear(UEINTX, RXOUTI);
	}
	if (bit_is_set(UEINTX, FIFOCON) && bit_is_clear(UEINTX,RWAL)) {
		bit_clear(UEINTX, FIFOCON);
	}
	UENUM = EP_MS_IN;
	if (bit_is_set(UEINTX,TXINI)) {
		bit_clear(UEINTX,TXINI);
		bit_clear(UEINTX,NAKINI);
	}
	if (bit_is_set(UEINTX, FIFOCON)
		&& (bit_is_clear(UEINTX,RWAL) || bit_is_set(GPIOR2,USB162_RX_SEND_REQ2))) {
		bit_clear(GPIOR2, USB162_RX_SEND_REQ2);
		bit_clear(UEINTX, FIFOCON);
	}
	usbPoll();
}

unsigned char usbmsc_getc(void) {
	return __usbmsc_getc();
}

void usbmsc_putc(unsigned char data) {
	__usbmsc_putc(data);
}

void usbmsc_putc2(uint8_t ch) {
	usbmsc_putc(0); 
	usbmsc_putc(ch); 
}

void usbmsc_putc4(uint8_t ch) {
	usbmsc_putc2(0); 
	usbmsc_putc2(ch); 
}

void usbmsc_putc2_same(uint8_t ch) {
	usbmsc_putc(ch); 
	usbmsc_putc(ch); 
}


static inline void usbmsc_oflush() {
	bit_set(GPIOR2, USB162_RX_SEND_REQ2);
}

/* ---- */

#define MSC_LOG
#ifndef MSC_LOG
#define msc_debug(ch)
#else
#define MSC_LOG_SIZE	0xc0

static uint8_t msc_log[MSC_LOG_SIZE];
static uint8_t msc_log_idx;

static void __msc_debug(uint8_t ch) {
	if (msc_log_idx < MSC_LOG_SIZE) {
		msc_log[msc_log_idx++] = ch;

	}
}
#define msc_debug(ch)	__msc_debug(ch)
#endif

#define MSC_MAXLUN	2
void (* mscdev[MSC_MAXLUN])(uint8_t,uint8_t);
uint8_t msc_sense_key[MSC_MAXLUN+1];
uint8_t msc_sense_asc[MSC_MAXLUN+1];
static uint8_t msc_maxlun;

#ifndef MSC_UNIT_NOT_READY
#define MSC_UNIT_NOT_READY		0
#define MSC_UNIT_CHANGE_TO_READY	1
#define MSC_UNIT_READY			2
#endif

uint8_t msc_unit_state[MSC_MAXLUN];

msc_cbw_t msc_cbw;

#define MSC_RESULT_OK		0
#define MSC_RESULT_CMD_ERR	1
#define MSC_RESULT_FRAME_ERR	2

static uint8_t last_err_cmd;

volatile uint32u_t msc_trans_offset;

/*
#define MSC_TRANS_CONT	0
#define MSC_TRANS_DONE	1
*/

uint8_t (* msc_trans_proc)();

#define MSC_STATE_COMMAND	0
#define MSC_STATE_TRANS		1
#define MSC_STATE_STATUS	2

static uint8_t msc_state = MSC_STATE_COMMAND;


#define MSC_SENSE_PAGE_CODE_ERR_6	(1<<0)
#define MSC_SENSE_PAGE_CODE_ERR_10	(1<<1)
#define MSC_SENSE_PAGE_CODE_EXPT_6	(1<<2)
#define MSC_SENSE_PAGE_CODE_EXPT_10	(1<<3)
#define MSC_SENSE_PAGE_CODE_CACHING	(1<<4)

static uint8_t msc_sense_mode;
static uint8_t msc_pagecode_len;
static uint8_t msc_write_protected;

uint8_t msc_lock_stat;

int8_t msc_devattach(void (*func)(uint8_t,uint8_t)) {
	int8_t ret;
	if (msc_maxlun >= MSC_MAXLUN) 
		return -1;
	ret = msc_maxlun;
	mscdev[msc_maxlun++] = func;
	return ret;
}

static inline void msc_devcall(uint8_t lun, uint8_t cmd) {
	mscdev[lun](lun,cmd);
}

void callback_msc_reset(uint8_t hard) {
	int8_t i;
	/* endpoint FIFO Reset */
	bit_set(UERST , EP_MS_IN);
	bit_clear(UERST , EP_MS_IN);
	bit_set(UERST , EP_MS_OUT);
	bit_clear(UERST , EP_MS_OUT);
	if (hard) {
		msc_debug(0xfe);
	} else {
		msc_debug(0xfd);
	}
	msc_state = MSC_STATE_COMMAND;
	for (i=0; i<msc_maxlun; i++) {
		msc_unit_state[i] = MSC_UNIT_NOT_READY;
	}
}

uint8_t callback_msc_maxlun() {
	msc_debug(0xfc);
	return (msc_maxlun==0)?0:msc_maxlun-1;
}

uint8_t msc_trans_skip() {
	if (msc_cbw.cbw.trans_len.u32 == 0L) return MSC_TRANS_DONE;
#if 0
	if (msc_cbw.cbw.flags & 0x80) { /* Bulk IN */
	    if (usbmsc_can_putc(1)) {
		while (usbmsc_can_putc(1) && (msc_cbw.cbw.trans_len.u32 > 0L)) {
			usbmsc_putc(0);
			msc_cbw.cbw.trans_len.u32--;
		}
	    }
	} else { /* Bulk OUT */
		while (usbmsc_can_getc(1) && (msc_cbw.cbw.trans_len.u32 > 0L)) {
			usbmsc_getc(); 
			msc_cbw.cbw.trans_len.u32--;
		}
	}
#else
	while (msc_cbw.cbw.trans_len.u32 > 0L) {
	    if (msc_cbw.cbw.flags & 0x80) { /* Bulk IN */
		if (!usbmsc_can_putc(1))
			break;
		usbmsc_putc(0);
	    } else {
		if (!usbmsc_can_getc(1))
			break;
		usbmsc_getc(); 
	    }
	    msc_cbw.cbw.trans_len.u32--;
	}
#endif
	return MSC_TRANS_CONT;
}

static uint8_t msc_trans_request_sense() {
	uint8_t sc_lun = msc_cbw.cbw.lun;
	if (usbmsc_can_putc(18)) {
		usbmsc_putc(0x70); // RESPONSE_CODE_SENSE
		usbmsc_putc(0x00);
		usbmsc_putc(msc_sense_key[sc_lun]);
		usbmsc_putc4(0x00);
		usbmsc_putc(10); // ADDITIONAL_SENSE_LENGTH
		usbmsc_putc4(0x00);
		usbmsc_putc(msc_sense_asc[sc_lun]);
		usbmsc_putc(0x00); // msc_sense_ascq
		usbmsc_putc4(0x00);
		msc_cbw.cbw.trans_len.ub[0] -= 18;
		msc_sense_key[sc_lun] = SC_SENSE_KEY_NO_SENSE;
		msc_sense_asc[sc_lun] = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION;
		msc_trans_proc = msc_trans_skip;
	}
	return MSC_TRANS_CONT;
}

static uint8_t msc_trans_mode_sense() {
	if (usbmsc_can_putc(32)) {
		usbmsc_putc(msc_pagecode_len); // Data length
		usbmsc_putc(0x00); // DeviceType: DEFAULT_MEDIUM_TYPE
		usbmsc_putc(msc_write_protected ? 0x80:0x00);
		usbmsc_putc(0x00); // BLOCK_DESCRIPTOR_LENGTH
		msc_cbw.cbw.trans_len.ub[0] -= 4;

	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_ERR_6) {
		usbmsc_putc(SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY);
		usbmsc_putc(6);    // PAGE_LENGTH
		usbmsc_putc(0x80);
		usbmsc_putc(0x03);  // READ_RETRY_COUNT
		usbmsc_putc4(0x00); // CORRECTION_SPAN,HEAD_OFFSET_COUNT
				    // ,DATA_STROBE_OFFSET,Reserved
		msc_cbw.cbw.trans_len.ub[0] -= 8;
	    } else if (msc_sense_mode & MSC_SENSE_PAGE_CODE_ERR_10) {
		usbmsc_putc(SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x80);
		usbmsc_putc(0x03);  // READ_RETRY_COUNT
		usbmsc_putc4(0x00); // CORRECTION_SPAN,HEAD_OFFSET_COUNT
				    // ,DATA_STROBE_OFFSET,Reserved

		usbmsc_putc(0x80);  // WRITE_RETRY_COUNT
		usbmsc_putc(0x00);
		usbmsc_putc2(0x00); // RECOVERY_LIMIT
		msc_cbw.cbw.trans_len.ub[0] -= 12;
	    }
	    if (msc_sense_mode & MSC_SENSE_PAGE_CODE_EXPT_6) {
		usbmsc_putc(SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS);
		usbmsc_putc(6);    // PAGE_LENGTH
		usbmsc_putc(0x00);
		usbmsc_putc(0x05);  // MRIE
		usbmsc_putc4(0x00);
		msc_cbw.cbw.trans_len.ub[0] -= 8;
	    }
#if 0
	    else if (msc_sense_mode & MSC_SENSE_PAGE_CODE_EXPT_10) {
		// Page Code: Informational exceptions control page
		// See chapter 8.3.8 on SPC-2 specification
		usbmsc_putc(SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x00);
		usbmsc_putc(0x05);  // MRIE
		usbmsc_putc4(0x00); // Interval Timer 
		usbmsc_putc4(0x01); // Report Count
		msc_cbw.cbw.trans_len.ub[0] -= 12;
#endif
#if 0
	    } else if (msc_sense_mode & MSC_SENSE_PAGE_CODE_CACHING) {
		usbmsc_putc(SC_PAGE_CODE_CACHING_MODE_PAGE);
		usbmsc_putc(10);    // PAGE_LENGTH
		usbmsc_putc(0x04);  // Write Cache Enable , Read Cache Disabled
		usbmsc_putc(0x00);
		usbmsc_putc2_same(0xff); // don't disable prefetch
		usbmsc_putc2(0x00);
		usbmsc_putc2_same(0xff); //  Maximum prefetch
		usbmsc_putc2_same(0xff); // Maximum prefetch ceiling
		msc_cbw.cbw.trans_len.ub[0] -= 12;
	    }
#endif
	    msc_trans_proc = msc_trans_skip;
	}
	return MSC_TRANS_CONT;
}

#endif

#define USE_SERFLASH
#ifdef USE_SERFLASH
extern void msc_serflash_init();
#endif

void msc_init() {
#ifdef EP_MS_OUT
#ifdef USE_SERFLASH
	msc_serflash_init();
#endif
#endif
}

void msc_print_status() {
#if defined(EP_TX) && defined(EP_MS_OUT)
#ifdef MSC_LOG
	uint8_t i,j,d;
	uint16_t offset = 0;
	usbcdc_put_pstr(PSTR("msc "));
	usbcdc_putc('0' + msc_state);
	usbcdc_put_pstr(PSTR(" err:"));
	usbcdc_put_hex( last_err_cmd );
	usbcdc_put_pstr(PSTR("\r\n"));
	for(i=0; i<12; i++) {
		d = (offset >> 8) & 0xff;
		usbcdc_put_hex(d);
		d = offset & 0xff;
		usbcdc_put_hex(d);
		usbcdc_putc(':');
		for (j=0; j<16; j++) {
			usbcdc_putc(' ');
			d = msc_log[offset];
			usbcdc_put_hex(d);
			offset++;
		}
		usbcdc_put_pstr(PSTR("\r\n"));
	}
	msc_log_idx = 0;
	memset(msc_log,0,MSC_LOG_SIZE);
#endif
#endif
}

void msc_poll() {
#ifdef EP_MS_OUT
	uint8_t i;
	uint8_t sc_cmd;
	uint8_t sc_len = 14;
	uint8_t sc_lun = msc_cbw.cbw.lun;
	uint8_t large_trans;
	uint8_t pagecode;

	switch (msc_state) {
	case MSC_STATE_COMMAND:
		if (msc_is_locked(MSC_OWNER_DRIVER)) return;
		if (usbmsc_can_getc(31)) {
			msc_lock(MSC_OWNER_DRIVER);
			for (i=0; i<15; i++) {
				msc_cbw.ub[i] = usbmsc_getc();
			}
			msc_trans_proc = msc_trans_skip;
			msc_trans_offset.u32 = 0L;
			sc_cmd = usbmsc_getc();
			//sc_lun = usbmsc_getc() >> 5;
			usbmsc_getc(); // skip lun
			msc_debug(0xc0 + sc_lun);
			msc_debug(sc_cmd);
			msc_debug(msc_cbw.cbw.trans_len.ub[0]);
			large_trans = msc_cbw.cbw.trans_len.ub[1] 
				    | msc_cbw.cbw.trans_len.ub[2] 
			            | msc_cbw.cbw.trans_len.ub[3];
			if (sc_lun >= msc_maxlun) {
				sc_lun = msc_cbw.cbw.lun = msc_maxlun;
				msc_sense_key[sc_lun] = SC_SENSE_KEY_NOT_READY;
				msc_sense_asc[sc_lun] = SC_ASC_LOGICAL_UNIT_NOT_READY;
			}
			if (sc_cmd != SC_REQUEST_SENSE) {
				msc_sense_key[sc_lun] = SC_SENSE_KEY_NO_SENSE;
				msc_sense_asc[sc_lun] = SC_ASC_NO_ADDITIONAL_SENSE_INFORMATION;
			}
			switch (sc_cmd) {
			case SC_INQUIRY:
			case SC_READ_CAPACITY:
				if (large_trans) goto ill_req;
				msc_devcall(sc_lun, sc_cmd);
				break;
			case SC_MODE_SENSE_6:
				if (large_trans) goto ill_req;
				pagecode = usbmsc_getc(); // pagecode
				sc_len -= 1;
				msc_debug(pagecode);
				msc_trans_proc = msc_trans_mode_sense;
				msc_sense_mode = 0;
				msc_pagecode_len = 3;
				if (msc_cbw.cbw.trans_len.ub[0] < 12) {
					break;
				}
				switch ( pagecode & SC_MSK_PAGE_CODE ) {
				case SC_PAGE_CODE_READ_WRITE_ERROR_RECOVERY:
					if (msc_cbw.cbw.trans_len.ub[0] < 16) {
						msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_6;
						msc_pagecode_len += 8;
						break;
					}
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_10;
					msc_pagecode_len += 12;
					break;
#if 0
				case SC_PAGE_CODE_CACHING_MODE_PAGE:
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_CACHING;
					msc_pagecode_len += 12;
					break;
#endif
				case SC_PAGE_CODE_INFORMATIONAL_EXCEPTIONS: 
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT_6;
					//msc_pagecode_len += 12;
					msc_pagecode_len += 8;
					break;
				case SC_PAGE_CODE_ALL:
					if (msc_cbw.cbw.trans_len.ub[0] < 16) {
						msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_6;
						msc_pagecode_len += 8;
						break;
					}
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_ERR_10;
					msc_pagecode_len += 12;
					if (msc_cbw.cbw.trans_len.ub[0] < 28) {
						break;
					}
					msc_sense_mode |= MSC_SENSE_PAGE_CODE_EXPT_6;
					msc_pagecode_len += 12;
					break;
				default:
					goto ill_req;
					//break;
				}
				msc_devcall(sc_lun, sc_cmd);
				break;
			case SC_PREVENT_ALLOW_MEDIUM_REMOVAL:
				break;
			case SC_READ_10:
			case SC_WRITE_10:
				msc_trans_offset.ub[3] = usbmsc_getc();
				msc_trans_offset.ub[2] = usbmsc_getc();
				msc_trans_offset.ub[1] = usbmsc_getc();
				msc_trans_offset.ub[0] = usbmsc_getc();
				sc_len -= 4;
				msc_cbw.cbw.trans_len.u32 >>= 5; // div 32
				if (msc_cbw.cbw.trans_len.u16[1]) 
					goto ill_req; // too large
				msc_devcall(sc_lun, sc_cmd);
				break;
			case SC_REQUEST_SENSE:
				if (large_trans) goto ill_req;
				msc_trans_proc = msc_trans_request_sense;
				break;
			case SC_TEST_UNIT_READY:
				msc_devcall(sc_lun, sc_cmd);
				if (msc_unit_state[sc_lun] == MSC_UNIT_NOT_READY) {
					msc_sense_key[sc_lun] = SC_SENSE_KEY_NOT_READY;
					msc_sense_asc[sc_lun] = SC_ASC_LOGICAL_UNIT_NOT_READY;
				} else if (msc_unit_state[sc_lun] == MSC_UNIT_CHANGE_TO_READY) {
					msc_sense_key[sc_lun] = SC_SENSE_KEY_UNIT_ATTENTION;
					msc_sense_asc[sc_lun] = SC_ASC_NOT_READY_TO_READY_CHANGE;
					msc_unit_state[sc_lun] = MSC_UNIT_READY;
				}
				break;
			case SC_VERIFY_10:
				break;
			case SC_READ_FORMAT_CAPACITY:
			default:
			ill_req:
				last_err_cmd = sc_cmd;
				msc_sense_key[sc_lun] = SC_SENSE_KEY_ILLEGAL_REQUEST;
				msc_sense_asc[sc_lun] = SC_ASC_INVALID_COMMAND_OPERATION_CODE;
			}
			for (i=0; i<sc_len; i++) {
				usbmsc_getc();
			}
			if (msc_cbw.cbw.trans_len.u32 == 0L) {
				msc_state = MSC_STATE_STATUS;
			} else {
				msc_state = MSC_STATE_TRANS;
			}
		}
		break;
	case MSC_STATE_TRANS:
		if ((msc_trans_proc)() == MSC_TRANS_DONE) {
		    if (msc_cbw.cbw.flags & 0x80) { // Bulk IN
			usbmsc_oflush();
		    }
		    msc_state = MSC_STATE_STATUS;
		}
		break;
	case MSC_STATE_STATUS:
		if (usbmsc_can_putc(13)) {
			msc_cbw.csw.magic[3] = 'S'; // 'USBC' -> 'USBS'
			msc_cbw.csw.status = 
				(msc_sense_key[msc_cbw.cbw.lun] == SC_SENSE_KEY_NO_SENSE)?0:1;
			for (i=0; i<13; i++) {
				usbmsc_putc(msc_cbw.ub[i]); 
			}
			msc_debug(msc_sense_key[msc_cbw.cbw.lun]);
			msc_state = MSC_STATE_COMMAND;
			usbmsc_oflush();
		}
		msc_unlock();
		break;
	default:
		break;
	}
	usbmsc_poll();
#endif
}
