/* Name: cdc162.c
 * Project: AVR USB driver for AT90USB
 * Author: Koji Suzuki
 * Creation Date: 2007-12-9
 * Tabsize: 8
 * Copyright: (c) 2007 by Koji Suzuki (suz@nmj2.sakura.ne.jp)
 * License: Proprietary, free under certain conditions. See Documentation.
 */

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))

#define nop()	__asm__ volatile ("nop" )
#define	wdr()	__asm__ volatile("wdr" )

#include "usbconfig.h"
#include "usb162.h"
#include "cdc162.h"
#include "msc162.h"
#include "serflash.h"

#define SC_VENDOR_ID	'U','S','B','1','6','2',' ',' '
#define SC_PRODUCT_ID	'S','E','R','F','L','A','S','H',' ',' ',' ',' ',' ',' ',' ',' '
#define SC_REVISION_ID	'0','.','0','2'

static uint8_t serflash_inquiry_data[] PROGMEM = {
	0x00, // DeviceType: Direct-access device , 0x05: CDROM
	0x80, // Medium is removable
	0x03, // Version : SPC (=SBC) (ANSI SCSI level 3 ?)
	0x02, // Response data format
	31,   // Additional Length
	0,0,0, // Reserved
	SC_VENDOR_ID,
	SC_PRODUCT_ID,
	SC_REVISION_ID,
};

#define WRITE_STAT_INIT		0
#define WRITE_STAT_PGM		1
#define WRITE_STAT_ERASE	2

static uint8_t write_stat;

static uint8_t serflash_trans_inquiry() {
	while (usbmsc_can_putc(1) && (msc_cbw.cbw.trans_len.ub[0] > 0)) {
		usbmsc_putc(__LPM(serflash_inquiry_data+msc_trans_offset.ub[0]));
		msc_trans_offset.ub[0]++;
		msc_cbw.cbw.trans_len.ub[0] --;
		if (msc_trans_offset.ub[0] >= 36) {
	    		msc_trans_proc = msc_trans_skip;
			break;
		}
	}
	return MSC_TRANS_CONT;
}

static uint8_t serflash_trans_read_capacity() {
	uint32_t capacity;
	capacity = serflash_size - 1;

	if (usbmsc_can_putc(8)) {
		usbmsc_putc((uint8_t)(capacity>>24));
		usbmsc_putc((uint8_t)(capacity>>16));
		usbmsc_putc((uint8_t)(capacity>>8));
		usbmsc_putc((uint8_t)(capacity));
		usbmsc_putc2(0x00); // SECTOR SIZE
		usbmsc_putc(MSC_SECTOR_SIZE >> 8);
		usbmsc_putc(MSC_SECTOR_SIZE & 0xff);
		msc_cbw.cbw.trans_len.ub[0] -= 8;
		msc_trans_proc = msc_trans_skip;
	}
	return MSC_TRANS_CONT;
}

static uint8_t serflash_trans_read() {
	uint8_t i;
	while (usbmsc_can_putc(32) && (msc_cbw.cbw.trans_len.u16[0] > 0)) {
#ifdef HARD_SPI2
		spi2_send(0);
		for (i=0; i<31; i++) {
			__usbmsc_putc_nc(spi2_recv_send(0));
		}
		__usbmsc_putc_nc(spi2_recv(0));
#else
		for (i=0; i<32; i++) {
			__usbmsc_putc_nc(serflash_proc(0));
		}
#endif
		msc_cbw.cbw.trans_len.u16[0]--;
	}
	if (msc_cbw.cbw.trans_len.u16[0] == 0) {
		serflash_close();
		return MSC_TRANS_DONE;
	}
	return MSC_TRANS_CONT;
}

static uint8_t serflash_trans_write() {
	uint8_t i;
	while (usbmsc_can_getc(32) && (msc_cbw.cbw.trans_len.u16[0] > 0L)) {
		if (msc_trans_offset.ub[0] == 0) {
			serflash_close();
			if (serflash_status() & SERFLASH_STAT_WIP) {
				return MSC_TRANS_CONT;
			}
#ifdef SUPPORT_ERASE4K
			if ((msc_trans_offset.ub[1] & 0x0f) == 0) {
				if (write_stat != WRITE_STAT_ERASE) {
					write_stat = WRITE_STAT_ERASE;
					serflash_erase_4k(msc_trans_offset.u32);
					return MSC_TRANS_CONT;
				}
			}
#else
			if (msc_trans_offset.ub[1] == 0) {
				if (write_stat != WRITE_STAT_ERASE) {
					write_stat = WRITE_STAT_ERASE;
					serflash_erase_64k(msc_trans_offset.u32);
					return MSC_TRANS_CONT;
				}
			}
#endif
			write_stat = WRITE_STAT_PGM;
			serflash_write_enable();
			serflash_open(SERFLASH_PAGE_PGM,msc_trans_offset.u32);
		}
#ifdef HARD_SPI2
		spi2_send(__usbmsc_getc_nc()); 
		for (i=0; i< 31; i++) {
			spi2_recv_send(__usbmsc_getc_nc()); 
		}
		spi2_recv(); 
#else
		for (i=0; i< 32; i++) {
			serflash_proc(__usbmsc_getc_nc()); 
		}
#endif
		msc_cbw.cbw.trans_len.u16[0] --;
		msc_trans_offset.u32 += 32;
	}
	if (msc_cbw.cbw.trans_len.u16[0] == 0) {
		serflash_close();
		if (serflash_status() & SERFLASH_STAT_WIP) {
			return MSC_TRANS_CONT;
		}
		write_stat = WRITE_STAT_INIT;
		return MSC_TRANS_DONE;
	}
	return MSC_TRANS_CONT;
}

static void serflash_devcall(uint8_t sc_lun, uint8_t sc_cmd) {
	switch (sc_cmd) {
	case SC_INQUIRY:
		msc_trans_proc = serflash_trans_inquiry;
		break;
	case SC_READ_CAPACITY:
		msc_trans_proc = serflash_trans_read_capacity;
		break;
	case SC_READ_10:
	case SC_WRITE_10:
		msc_trans_offset.u32 <<= MSC_SECTOR_SHIFT;
		if (sc_cmd == SC_READ_10) {
			serflash_open(SERFLASH_READ,msc_trans_offset.u32);
			msc_trans_proc = serflash_trans_read;
		} else {
			msc_trans_proc = serflash_trans_write;
		}
		break;
	case SC_TEST_UNIT_READY:
		if (msc_unit_state[sc_lun] == MSC_UNIT_NOT_READY) {
			if (serflash_type) {
				msc_unit_state[sc_lun] = MSC_UNIT_CHANGE_TO_READY;
			}
		}
		break;
	default:
		break;
	}
}

void msc_serflash_init() {
	serflash_init();
	msc_devattach(serflash_devcall);
}
